# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function
import sys
import re
from parglare.exceptions import GrammarError

if sys.version < '3':
    text = unicode  # NOQA
else:
    text = str

# Associativity
ASSOC_NONE = 0
ASSOC_LEFT = 1
ASSOC_RIGHT = 2

# Priority
DEFAULT_PRIORITY = 10


def escape(instr):
    return instr.replace('\n', r'\n').replace('\t', r'\t')


class GrammarSymbol(object):
    def __init__(self, name):
        self.name = escape(name)

    def __unicode__(self):
        return str(self)

    def __str__(self):
        return self.name

    def __repr__(self):
        return "{}({})".format(type(self).__name__, str(self))

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        return self.name == other.name


class NonTerminal(GrammarSymbol):
    pass


class Terminal(GrammarSymbol):
    def __init__(self, name):
        self.prior = DEFAULT_PRIORITY
        super(Terminal, self).__init__(name)


class TerminalStr(Terminal):
    def __init__(self, name, value=None):
        super(TerminalStr, self).__init__(name)
        self.value = value if value else name

    def parse(self, in_str, pos):
        if in_str[pos:pos+len(self.value)] == self.value:
            return self.value
        else:
            return ''


class TerminalRegEx(Terminal):
    def __init__(self, name, regex=None):
        super(TerminalRegEx, self).__init__(name)
        self._regex = regex if regex else name
        self.regex = re.compile(self._regex)

    def parse(self, in_str, pos):
        m = self.regex.match(in_str, pos)
        if m:
            matched = m.group()
            return matched
        else:
            return ''


class TerminalEmpty(Terminal):
    def parse(self, in_str, pos):
        return ''


class TerminalEOF(Terminal):
    def parse(self, in_str, pos):
        return ''


class TerminalStop(Terminal):
    def parse(self, in_str, pos):
        return ''


# These two terminals are special terminals used internally.
AUGSYMBOL = NonTerminal("S'")
STOP = TerminalStop("STOP")

# These two terminals are special terminals used in the grammars.
# EMPTY will match nothing and always succeeds.
# EOF will match only at the end of the input string.
EMPTY = TerminalEmpty("EMPTY")
EOF = TerminalEOF("EOF")


class Production(object):
    """Represent production from the grammar.

    Attributes:
    symbol (GrammarSymbol):
    rhs (ProductionRHS):
    assoc (int): Associativity. Used for ambiguity (shift/reduce) resolution.
    prior (int): Priority. Used for ambiguity (shift/reduce) resolution.
    prod_id (int): Ordinal number of the production.
    prod_symbol_id (str): Ident of production in the form "<Symbol>:ord" where
        ordinal is ordinal of the alternative choice starting from 0. Used to
        map actions.
    """

    def __init__(self, symbol, rhs, assoc=ASSOC_NONE, prior=DEFAULT_PRIORITY):
        """
        Args:
        symbol (GrammarSymbol): A grammar symbol on the LHS of the production.
        rhs (list of GrammarSymbols):
        """
        self.symbol = symbol
        self.rhs = rhs if rhs else ProductionRHS()
        self.assoc = assoc
        self.prior = prior

    def __str__(self):
        if hasattr(self, 'prod_id'):
            return "%d: %s = %s" % (self.prod_id, self.symbol, self.rhs)
        else:
            return "%s = %s" % (self.symbol, self.rhs)

    def __repr__(self):
        return 'Production({})'.format(str(self.symbol))


class ProductionRHS(list):
    def __getitem__(self, key):
        try:
            return super(ProductionRHS, self).__getitem__(key)
        except IndexError:
            return None

    def __str__(self):
        return " ".join([str(x) for x in self])


class Grammar(object):
    """
    Grammar is a collection of production rules.
    First production is reserved for the augmented production (S' -> S).

    Attributes:
    productions (list of Production):
    root_symbol (GrammarSymbol): start/root symbol of the grammar.
    nonterminals (set of NonTerminal):
    terminals(set of Terminal):

    """

    def __init__(self, productions, root_symbol=None):
        self.productions = productions
        self.root_symbol = \
            root_symbol if root_symbol else productions[0].symbol
        self.init_grammar()

    def init_grammar(self):
        """
        Extracts all grammar symbol (nonterminal and terminal) from the
        grammar and check references in productions.
        """
        self.nonterminals = set()
        self.terminals = set()

        # Reserve 0 production. It is used for augmented prod. in LR
        # automata calculation.
        self.productions.insert(
            0,
            Production(AUGSYMBOL, ProductionRHS([self.productions[0].symbol,
                                                 STOP])))

        for s in self.productions:
            if isinstance(s.symbol, NonTerminal):
                self.nonterminals.add(s.symbol)
            else:
                raise GrammarError("Invalid production symbol '%s' "
                                   "for production '%s'" % (s.symbol,
                                                            text(s)))
            for idx, t in enumerate(s.rhs):
                if isinstance(t, Terminal):
                    self.terminals.add(t)
                elif isinstance(t, text):
                    term = TerminalStr(t)
                    self.terminals.add(term)
                    s.rhs[idx] = term

        # Enumerate productions
        idx_per_symbol = {}
        for idx, s in enumerate(self.productions):
            s.prod_id = idx
            s.prod_symbol_id = "{}:{}".format(s.symbol,
                                              idx_per_symbol.get(s.symbol, 0))
            idx_per_symbol[s.symbol] = idx_per_symbol.get(s.symbol, 0) + 1
            for ref in s.rhs:
                if ref not in self.nonterminals and ref not in self.terminals:
                    raise GrammarError("Undefined grammar symbol '%s' "
                                       "referenced from production '%s'."
                                       % (ref, s))

    def get_terminal(self, name):
        "Returns terminal with the given name."
        for t in self.terminals:
            if t.name == name:
                return t

    def get_nonterminal(self, name):
        "Returns non-terminal with the given name."
        for n in self.nonterminals:
            if n.name == name:
                return n

    def get_symbol(self, name):
        "Returns grammar symbol with the given name."
        s = self.get_terminal(name)
        if not s:
            s = self.get_nonterminal(name)
        return s

    def get_production(self, name):
        "Returns first production number for the given symbol name"
        for idx, p in enumerate(self.productions):
            if p.symbol.name == name:
                return idx

    @staticmethod
    def from_string(grammar_str):
        from parglare import Parser
        global pg_productions, GRAMMAR, pg_actions
        p = Parser(create_grammar(pg_productions, GRAMMAR), actions=pg_actions)
        prods = p.parse(grammar_str)
        return Grammar(prods)

    @staticmethod
    def from_file(file_name):
        from parglare import Parser
        global pg_productions, GRAMMAR, pg_actions
        p = Parser(create_grammar(pg_productions, GRAMMAR), actions=pg_actions)
        prods = p.parse_file(file_name)
        return Grammar(prods)

    def print_debug(self):
        print("Terminals:")
        print(" ".join([text(t) for t in self.terminals]))
        print("NonTerminals:")
        print(" ".join([text(n) for n in self.nonterminals]))

        print("Productions:")
        for p in self.productions:
            print(p)


def create_grammar(productions, start_symbol=None):
    """Creates grammar from a list of productions given in the form: (LHS, RHS).
    Where LHS is grammar symbol and RHS is a list or tuple of grammar symbols
    from the right-hand side of the production.
    """
    gp = []
    for p in productions:
        assoc = ASSOC_NONE
        prior = DEFAULT_PRIORITY
        if len(p) > 2:
            assoc = p[2]
        if len(p) > 3:
            prior = p[3]

        gp.append(Production(p[0], ProductionRHS(p[1]), assoc, prior))
    g = Grammar(gp, start_symbol)
    return g


# Grammar for grammars

(GRAMMAR,
 PRODUCTION_SET,
 PRODUCTION,
 PRODUCTION_RHSS,
 PRODUCTION_RHS,
 GSYMBOL,
 NONTERM_REF,
 TERM,
 ASSOC,
 ASSOC_PRIOR,
 SEQUENCE) = [NonTerminal(name) for name in [
    'Grammar',
    'ProductionSet',
    'Production',
    'ProductionRHSs',
    'ProductionRHS',
    'GSymbol',
    'NonTermRef',
    'Term',
    'Assoc',
    'AssocPrior',
    'Sequence']]
NAME = TerminalRegEx('Name', r'[a-zA-Z0-9]+')
STR_TERM = TerminalRegEx("StrTerm",
                         r'''(?s)('[^'\\]*(?:\\.[^'\\]*)*')|'''
                         r'''("[^"\\]*(?:\\.[^"\\]*)*")''')
REGEX_TERM = TerminalRegEx("RegExTerm",
                           r'''\/((\\/)|[^/])*\/''')
PRIOR = TerminalRegEx("Prior", r'\d+')
pg_productions = [
    [GRAMMAR, [PRODUCTION_SET]],
    [PRODUCTION_SET, [PRODUCTION]],
    [PRODUCTION_SET, [PRODUCTION_SET, PRODUCTION]],
    [PRODUCTION, [NAME, '=', PRODUCTION_RHSS, ';']],
    [PRODUCTION_RHS, [SEQUENCE]],
    [PRODUCTION_RHS, [SEQUENCE, '{', ASSOC_PRIOR, '}']],
    [PRODUCTION_RHSS, [PRODUCTION_RHS], ASSOC_LEFT, 5],
    [PRODUCTION_RHSS, [PRODUCTION_RHSS, '|', PRODUCTION_RHS], ASSOC_LEFT, 5],
    [ASSOC_PRIOR, [ASSOC]],
    [ASSOC_PRIOR, [PRIOR]],
    [ASSOC_PRIOR, [ASSOC_PRIOR, ',', ASSOC_PRIOR], ASSOC_LEFT],
    [ASSOC, ['left']],
    [ASSOC, ['right']],
    [SEQUENCE, [GSYMBOL]],
    [SEQUENCE, [SEQUENCE, GSYMBOL]],
    [GSYMBOL, [NONTERM_REF]],
    [GSYMBOL, [TERM]],
    [NONTERM_REF, [NAME]],
    [TERM, [STR_TERM]],
    [TERM, [REGEX_TERM]],
]


def act_term_str(_, nodes):
    value = nodes[0].value[1:-1]
    name = value
    value = value.replace(r'\"', '"')\
                 .replace(r"\'", "'")\
                 .replace(r"\\", "\\")\
                 .replace(r"\n", "\n")\
                 .replace(r"\t", "\t")
    return TerminalStr(name, value)


def act_term_regex(_, nodes):
    value = nodes[0].value[1:-1]
    return TerminalRegEx(value, value)


def act_sequence(_, nodes):
    res = nodes[0]
    res.append(nodes[1])
    return res


def act_prod_rhss(_, nodes):
    res = nodes[0]
    res.append(nodes[2])
    return res


def act_production(_, nodes):
    symbol = NonTerminal(nodes[0])
    prods = []
    for p_rhs in nodes[2]:
        asoc = ASSOC_NONE
        prior = DEFAULT_PRIORITY
        if len(p_rhs) > 1:
            asoc = p_rhs[1]
        if len(p_rhs) > 2:
            prior = p_rhs[2]
        prods.append(Production(symbol, p_rhs[0], asoc, prior))

    return prods


def act_prodset(_, nodes):
    res = nodes[0]
    res.extend(nodes[1])
    return res


def act_assoc_prior(_, nodes):
    res = []
    res.extend(nodes[0])
    res.extend(nodes[2])
    return res


def act_production_rhs(_, nodes):
    assoc = ASSOC_NONE
    prior = DEFAULT_PRIORITY
    for ap in nodes[2]:
        if type(ap) is int:
            prior = ap
        else:
            if ap == 'left':
                assoc = ASSOC_LEFT
            elif ap == 'right':
                assoc = ASSOC_RIGHT
    return [ProductionRHS(nodes[0]), assoc, prior]


def act_grammar(_, nodes):
    prods = nodes[0]

    # Remove terminal production rules but first replace its
    # references.
    terms = {}
    to_del = []
    for idx, p in enumerate(prods):
        if len(p.rhs) == 1 and isinstance(p.rhs[0], Terminal):
            # Optimization: If a production has a single terminal RHS and there
            # is only one production for this grammar symbol then treat is as a
            # terminal.
            if len([x for x in prods if x.symbol == p.symbol]) == 1 and \
               p.symbol.name != 'LAYOUT':
                t = p.rhs[0]
                terms[p.symbol.name] = t
                t.name = p.symbol.name
                t.prior = p.prior
                if p.assoc != ASSOC_NONE:
                    raise GrammarError(
                        'Terminals should not define associativity '
                        '(terminal "{}").'.format(t.name))
                assert p.assoc == ASSOC_NONE
                to_del.append(idx)

    for idx in sorted(to_del, reverse=True):
        del prods[idx]

    # Change terminal references
    for p in prods:
        for idx, ref in enumerate(p.rhs):
            if ref.name in terms:
                p.rhs[idx] = terms[ref.name]
            elif ref.name == 'EMPTY':
                p.rhs[idx] = EMPTY
            elif ref.name == 'EOF':
                p.rhs[idx] = EOF

    return prods


pg_actions = {
    "Assoc:0": lambda _, nodes: nodes[0].value,
    "Assoc:1": lambda _, nodes: nodes[0].value,
    "AssocPrior:0": lambda _, nodes: [nodes[0]],
    "AssocPrior:1": lambda _, nodes: [nodes[0]],
    "AssocPrior:2": act_assoc_prior,
    "Prior": lambda _, value: int(value),
    "Term:0": act_term_str,
    "Term:1": act_term_regex,
    "Name": lambda _, value: value,
    "NonTermRef": lambda _, nodes: NonTerminal(nodes[0]),
    "GSymbol": lambda _, nodes: nodes[0],
    "Sequence:0": lambda _, nodes: [nodes[0]],
    "Sequence:1": act_sequence,
    "ProductionRHS:0": lambda _, nodes: [ProductionRHS(nodes[0])],
    "ProductionRHS:1": act_production_rhs,
    "ProductionRHSs:0": lambda _, nodes: [nodes[0]],
    "ProductionRHSs:1": act_prod_rhss,
    "Production": act_production,
    "ProductionSet:0": lambda _, nodes: nodes[0],
    "ProductionSet:1": act_prodset,
    "Grammar": act_grammar
}



from __future__ import unicode_literals
import pytest
from os.path import join, dirname
from parglare import Parser, Grammar, SLR, LALR
from .expression_grammar import get_grammar
from parglare.exceptions import ShiftReduceConflict, ParseError


def test_parsing():
    grammar = get_grammar()
    p = Parser(grammar)
    assert p.parse("id+id+id")


def test_parsing_from_file():
    grammar = get_grammar()
    p = Parser(grammar)
    assert p.parse_file(join(dirname(__file__), 'parsing_from_file.txt'))


def test_lr_1_grammar():
    """From the Knuth's 1965 paper: On the Translation of Languages from Left to
    Right

    """
    grammar = """
    S = 'a' A 'd' | 'b' A B;
    A = 'c' A | 'c';
    B = 'd';
    """

    g = Grammar.from_string(grammar)
    parser = Parser(g)

    parser.parse("acccccccccd")
    parser.parse("bcccccccccd")


def test_slr_conflict():
    """
    Unambiguous grammar which is not SLR(1).
    From the Dragon Book.
    This grammar has a S/R conflict if SLR tables are used.
    """

    grammar = """
    S = L '=' R | R;
    L = '*' R | 'id';
    R = L;
    """

    grammar = Grammar.from_string(grammar)
    with pytest.raises(ShiftReduceConflict):
        Parser(grammar, tables=SLR)

    Parser(grammar, tables=LALR)


def test_lalr_reduce_reduce_conflict():
    """
    Naive merging of states can lead to R/R conflict as shown in this grammar
    from the Dragon Book.
    """

    grammar = """
    S = 'a' A 'd' | 'b' B 'd' | 'a' B 'e' | 'b' A 'e';
    A = C;
    B = C;
    C = 'c';
    """
    grammar = Grammar.from_string(grammar)
    Parser(grammar)


def test_partial_parse():
    """
    Not giving EOF at the end of the sequence enables parsing of the beginning
    of the input string.
    """
    grammar = """
    S = 'a' B;
    B = 'b';
    """
    g = Grammar.from_string(grammar)
    parser = Parser(g)

    # Parser should succesfuly parse 'ab' at the beggining.
    parser.parse('abc')

    # But if EOF is given it will match only at the end of the string,
    # thus, the whole string must be parsed in order for parsing to
    # succeed.
    grammar = """
    S = 'a' B EOF;
    B = 'b';
    """
    g = Grammar.from_string(grammar)
    parser = Parser(g)
    parser.parse('a b')
    with pytest.raises(ParseError):
        parser.parse('a b c')



import pytest
from parglare.parser import Parser
from .expression_grammar_numbers import get_grammar, E


def test_actions():

    def sum_act(_, nodes):
        return nodes[0] + nodes[2]

    def t_act(_, nodes):
        if len(nodes) == 3:
            return nodes[0] * nodes[2]
        else:
            return nodes[0]

    def parenthesses_act(_, nodes):
        return nodes[1]

    def pass_act(_, nodes):
        return nodes[0]

    grammar = get_grammar()
    actions = {"number": lambda _, value: float(value),
               # Check action for each alternative
               "E:0": sum_act,
               "E:1": pass_act,
               # Check symbol-level action
               "T": t_act,
               # Again action for each alternative
               "F:0": parenthesses_act,
               "F:1": pass_act
    }

    p = Parser(grammar, actions=actions)

    result = p.parse("""34.7+78*34 +89+
    12.223*4""")

    assert result == 34.7 + 78 * 34 + 89 + 12.223 * 4



import pytest
from parglare.parser import Parser
from .expression_grammar import get_grammar, E


def test_tree_str():

    grammar = get_grammar()
    p = Parser(grammar)

    res = p.parse("""id+  id * (id
    +id  )
    """)

    ts = res.tree_str()

    assert '+[18, +]' in ts
    assert ')[23, )]' in ts



import pytest
from parglare.parser import first, follow
from parglare import Grammar, NonTerminal, TerminalStr, EMPTY
from parglare.grammar import STOP
from .expression_grammar import OPEN, ID, T, E, MULT, CLOSE, PLUS, get_grammar


def test_first():
    """
    Tests FIRST function.

    FIRST function returns a set of terminals that could start the sentence
    derived from the given non-terminal.
    """

    expression_grammar = get_grammar()
    first_set = first(expression_grammar)

    assert OPEN in first_set[T]
    assert ID in first_set[T]


def test_first_empty_in_rhs():
    """
    Test FIRST calculation when there are empty derivations in RHS of a
    production.
    """

    grammar = """
    S = A C;
    A = B | EMPTY;
    B = "b";
    C = "c";
    """

    g = Grammar.from_string(grammar)

    first_set = first(g)

    assert EMPTY in first_set[NonTerminal('A')]
    assert TerminalStr('B') in first_set[NonTerminal('A')]

    assert TerminalStr('B') in first_set[NonTerminal('S')]

    # 'A' can derive empty, thus 'C' must be in firsts of 'S'.
    assert TerminalStr('C') in first_set[NonTerminal('S')]


def test_follow():
    """Tests FOLLOW function.

    FOLLOW function calculates a set of terminals that can follow each grammar
    non-terminal in any of derived sentential forms.
    """

    expression_grammar = get_grammar()
    follow_set = follow(expression_grammar)

    assert follow_set[E] == set([CLOSE, PLUS, STOP])

    # Follow of T must contain all of follow of E
    assert follow_set[T] == set([MULT, CLOSE, PLUS, STOP])



import pytest
from parglare import Parser, Grammar


def assert_false(_, __):
    assert False


def test_priority():
    """
    Terminal priority is first checked. If there is multiple terminals
    that can match at given location, the one with the highest priority
    will be used.
    """

    grammar = """
    S = First | Second | Third;
    First = /\d+\.\d+/;
    Second = '14';
    Third = /\d+/ {15};
    """

    called = [False]

    def act_called(_, __):
        called[0] = True

    g = Grammar.from_string(grammar)
    actions = {
        "First": assert_false,
        "Second": assert_false,
        "Third": act_called,
    }
    parser = Parser(g, actions=actions)

    parser.parse('14')

    assert called[0]


def test_longest_match():
    "If priorities are the same longest-match strategy is used."

    grammar = """
    S = First | Second | Third;
    First = /\d+\.\d+/;
    Second = /\d+/;
    Third = /\d+a/;
    """
    g = Grammar.from_string(grammar)

    called = [False]

    def act_called(_, __):
        called[0] = True

    actions = {
        "First": act_called,
        "Second": assert_false,
        "Third": assert_false,
    }
    parser = Parser(g, actions=actions)

    # "First" will be used as it can match longer string.
    parser.parse('45.67')
    assert called[0]

    called = [False]

    def act_called(_, __):
        called[0] = True

    actions = {
        "First": assert_false,
        "Second": assert_false,
        "Third": act_called
    }
    parser = Parser(g, actions=actions)

    # "Third" will be used as it can match longer string.
    parser.parse('45a')
    assert called[0]


def test_most_specific():
    """If multiple terminals are of the same priority and matches the same length
    strings we shall prefer the most specific, i.e. TerminalStr over
    TerminalRegEx.
    """

    grammar = """
    S = First | Second | Third;
    First = "45";
    Second = /\d+/;
    Third = /\d+\.\d+/;
    """

    called = [False]

    def act_called(_, __):
        called[0] = True

    g = Grammar.from_string(grammar)
    actions = {
        "First": act_called,
        "Second": assert_false,
        "Third": assert_false,
    }
    parser = Parser(g, actions=actions)

    # "First" will be used as it is the most specific.
    parser.parse('45')
    assert called[0]


def test_if_all_fails():
    """
    If all strategies doesn't result in single terminal, choose first.
    TODO: Maybe some kind of warning would be useful.
    """

    # In this grammar all three terminal rules could be applied for string "b56"
    # They are of the same priority, same length match and all are regexes.
    # In this case choose the first one.
    grammar = """
    S = First | Second | Third;
    First = /(a|b)\d+/;
    Second = /(a|b|c)\d+/;
    Third = /(a|b|c|d)\d+/;
    """

    called = [False]

    def act_called(_, __):
        called[0] = True

    g = Grammar.from_string(grammar)
    actions = {
        "First": act_called,
        "Second": assert_false,
        "Third": assert_false,
    }
    parser = Parser(g, actions=actions)

    parser.parse("b56")



from collections import OrderedDict
from parglare.parser import LRItem, LRState
from parglare import NonTerminal
from .grammar import ProductionRHS, AUGSYMBOL, ASSOC_LEFT, ASSOC_NONE, STOP
from .exceptions import ShiftReduceConflict, ReduceReduceConflict
from .parser import Action, SHIFT, REDUCE, ACCEPT, first, follow
from .closure import closure, LR_1


def create_tables(parser, itemset_type):

    first_sets = first(parser.grammar)
    follow_sets = follow(parser.grammar, first_sets)

    g = parser.grammar
    start_prod_symbol = g.productions[parser.start_production].symbol
    g.productions[0].rhs = ProductionRHS([start_prod_symbol, STOP])

    # Create a state for the first production (augmented)
    s = LRState(parser, 0, AUGSYMBOL,
                [LRItem(g.productions[0], 0, set())])

    state_queue = [s]
    state_id = 1

    while state_queue:
        # For each state calculate its closure first, i.e. starting from a
        # so called "kernel items" expand collection with non-kernel items.
        # We will also calculate GOTO and ACTIONS dicts for each state. These
        # dicts will be keyed by a grammar symbol.
        state = state_queue.pop(0)
        closure(state, itemset_type, first_sets)
        parser._states.append(state)

        # Each state has its corresponding GOTO and ACTION tables
        goto = OrderedDict()
        parser._goto.append(goto)
        actions = OrderedDict()
        parser._actions.append(actions)

        # To find out other states we examine following grammar symbols
        # in the current state (symbols following current position/"dot")
        # and group all items by a grammar symbol.
        state._per_next_symbol = OrderedDict()

        # Each production has a priority. But since productions are grouped
        # by grammar symbol that is ahead we take the maximal
        # priority given for all productions for the given grammar symbol.
        state._max_prior_per_symbol = {}

        for i in state.items:
            symbol = i.production.rhs[i.position]
            if symbol:
                state._per_next_symbol.setdefault(symbol, []).append(i)

                # Here we calculate max priorities for each grammar symbol to
                # use it for SHIFT/REDUCE conflict resolution
                prod_prior = i.production.prior
                old_prior = state._max_prior_per_symbol.setdefault(
                    symbol, prod_prior)
                state._max_prior_per_symbol[symbol] = max(prod_prior,
                                                          old_prior)

        # For each group symbol we create new state and form its kernel
        # items from the group items with positions moved one step ahead.
        for symbol, items in state._per_next_symbol.items():
            inc_items = [i.get_pos_inc() for i in items]
            maybe_new_state = LRState(parser, state_id, symbol, inc_items)
            target_state = maybe_new_state
            try:
                idx = parser._states.index(maybe_new_state)
                target_state = parser._states[idx]
            except ValueError:
                try:
                    idx = state_queue.index(maybe_new_state)
                    target_state = state_queue[idx]
                except ValueError:
                    pass

            # We've found a new state. Register it for later processing.
            if target_state is maybe_new_state:
                state_queue.append(target_state)
                state_id += 1
            else:
                # State with this kernel items already exists.
                if itemset_type is LR_1:
                    # LALR: Try to merge states.
                    if not merge_states(target_state, maybe_new_state):
                        target_state = maybe_new_state
                        state_queue.append(target_state)
                        state_id += 1

            # Create entries in GOTO and ACTION tables
            if isinstance(symbol, NonTerminal):
                # For each non-terminal symbol we create an entry in GOTO
                # table.
                goto[symbol] = target_state

            else:
                if symbol is STOP:
                    actions[symbol] = Action(ACCEPT)
                else:
                    # For each terminal symbol we create SHIFT action in the
                    # ACTION table.
                    actions[symbol] = Action(SHIFT, state=target_state)

    # For LR(1) itemsets refresh/propagate item's follows as the LALR
    # merging might change item's follow in previous states
    if itemset_type is LR_1:
        for state in parser._states:

            # First refresh current state's follows
            closure(state, LR_1, first_sets)

            # Propagate follows to next states. GOTO table keeps information
            # about states created from this state
            inc_items = [i.get_pos_inc() for i in state.items]
            for target_state in parser._goto[state.state_id].values():
                for next_item in target_state.kernel_items:
                    next_item.follow.update(
                        inc_items[inc_items.index(next_item)].follow)

    # Calculate REDUCTION entries in ACTION tables and resolve possible
    # conflicts.
    for state in parser._states:
        actions = parser._actions[state.state_id]

        for i in state.items:
            if i.is_at_end:
                # If the position is at the end then this item
                # would call for reduction but only for terminals
                # from the FOLLOW set of item (LR(1)) or the production LHS
                # non-terminal (LR(0)).
                if itemset_type is LR_1:
                    f_set = i.follow
                else:
                    f_set = follow_sets[i.production.symbol]
                for t in f_set:
                    if t in actions:
                        # Conflict! Try to resolve
                        act = actions[t]
                        if act.action is SHIFT:
                            # SHIFT/REDUCE conflict. Use assoc and priority to
                            # resolve
                            act_prior = state._max_prior_per_symbol[
                                act.state.symbol]
                            prod = i.production
                            if prod.prior == act_prior:
                                if prod.assoc == ASSOC_NONE:
                                    if parser.debug:
                                        parser.print_debug()
                                    raise ShiftReduceConflict(state,
                                                              act.state.symbol,
                                                              prod)
                                elif prod.assoc == ASSOC_LEFT:
                                    # Override with REDUCE
                                    actions[t] = Action(REDUCE, prod=prod)
                                # If associativity is right leave SHIFT
                                # action
                            elif prod.prior > act_prior:
                                # This item operation priority is higher =>
                                # override with reduce
                                actions[t] = Action(REDUCE, prod=prod)
                                # If priority of existing SHIFT action is
                                # higher then leave it instead

                        else:
                            # REDUCE/REDUCE conflict
                            # Try to resolve using priorities
                            assert act.prod != i.production
                            prod = i.production
                            if act.prod.prior == prod.prior:
                                if parser.debug:
                                    parser.print_debug()
                                raise ReduceReduceConflict(state,
                                                           t,
                                                           act.prod,
                                                           i.production)
                            elif prod.prior > act.prod.prior:
                                actions[t] = Action(REDUCE, prod=prod)

                    else:
                        actions[t] = Action(REDUCE, prod=i.production)

    if parser.debug:
        parser.print_debug()


def merge_states(old_state, new_state):
    """Try to merge new_state on old_state if possible. If not possible return
    False.
    """

    item_pairs = []
    for i in old_state.kernel_items:
        new_item = new_state.get_item(i)
        item_pairs.append((i, new_item))

    # Check if merging would result in R/R conflict
    check_set = set()
    for old, new in [x for x in item_pairs if x[0].is_at_end]:
        if old.follow.intersection(check_set) \
               or new.follow.intersection(check_set):
            return False
        check_set.update(old.follow)
        check_set.update(new.follow)

    for old, new in item_pairs:
        old.follow.update(new.follow)
    return True



import pytest
from parglare.grammar import Grammar, Terminal, NonTerminal, ASSOC_LEFT, \
    ASSOC_RIGHT, DEFAULT_PRIORITY
from parglare.exceptions import GrammarError


def test_undefined_grammar_symbol():
    "Tests that undefined grammar symbols raises errors."
    grammar = """
    S = A B;
    A = "a" | B;
    B = id;
    """
    with pytest.raises(GrammarError) as e:
        Grammar.from_string(grammar)

    assert 'Undefined grammar symbol' in str(e)
    assert 'B = id' in str(e)


def test_terminal_nonterminal():

    # Production A is a terminal ("a") and non-terminal at the same time.
    # Thus, it must be recognized as non-terminal.
    grammar = """
    S = A B;
    A = "a" | B;
    B = "b";
    """
    g = Grammar.from_string(grammar)
    assert NonTerminal("A") in g.nonterminals
    assert Terminal("A") not in g.terminals
    assert Terminal("B") in g.terminals
    assert NonTerminal("B") not in g.nonterminals

    # Here A shoud be non-terminal while B will be terminal.
    grammar = """
    S = A B;
    A = B;
    B = "b";
    """

    g = Grammar.from_string(grammar)
    assert NonTerminal("A") in g.nonterminals
    assert Terminal("A") not in g.terminals
    assert Terminal("B") in g.terminals
    assert NonTerminal("B") not in g.nonterminals


def test_multiple_terminal_definition():

    # A is defined multiple times as terminal thus it must be recognized
    # as non-terminal with alternative expansions.
    grammar = """
    S = A A;
    A = "a";
    A = "b";
    """

    g = Grammar.from_string(grammar)

    assert NonTerminal("A") in g.nonterminals
    assert Terminal("A") not in g.terminals


def test_assoc_prior():
    """Test that associativity and priority can be defined for productions and
    terminals.
    """

    grammar = """
    E = E '+' E {left, 1};
    E = E '*' E {2, left};
    E = E '^' E {right};
    E = id;
    id = /\d+/;
    """

    g = Grammar.from_string(grammar)

    assert g.productions[1].prior == 1
    assert g.productions[1].assoc == ASSOC_LEFT
    assert g.productions[3].assoc == ASSOC_RIGHT
    assert g.productions[3].prior == DEFAULT_PRIORITY

    assert g.productions[3].prior == DEFAULT_PRIORITY


def test_terminal_priority():
    "Terminals might define priority which is used for lexical disambiguation."

    grammar = """
    S = A | B;
    A = 'a' {15};
    B = 'b';
    """

    g = Grammar.from_string(grammar)

    for t in g.terminals:
        if t.name == 'A':
            assert t.prior == 15
        else:
            assert t.prior == DEFAULT_PRIORITY


def test_no_terminal_associavitity():
    "Tests that terminals can't have associativity defined."
    grammar = """
    S = A | B;
    A = 'a' {15, left};
    B = 'b';
    """

    with pytest.raises(GrammarError) as e:
        Grammar.from_string(grammar)

    assert 'associativity' in str(e)



import pytest
from parglare import Parser, Grammar
from parglare.exceptions import ParseError
from .expression_grammar import get_grammar


def test_default_whitespaces():

    grammar = get_grammar()
    p = Parser(grammar)

    p.parse("""id+  id * (id
    +id  )
    """)


def test_whitespace_redefinition():

    grammar = get_grammar()

    # Make newline treated as non-ws characted
    p = Parser(grammar, ws=' \t')

    p.parse("""id+  id * (id +id  ) """)

    try:
        p.parse("""id+  id * (id
        +id  )
        """)
    except ParseError as e:
        assert e.position == 13


def test_whitespace_not_used_if_layout():
    """
    If LAYOUT rule is used, ws definition is ignored.
    """
    grammar = """
    S = 'a' 'b';
    LAYOUT = 'k' | EMPTY;
    """
    g = Grammar.from_string(grammar)
    parser = Parser(g)
    with pytest.raises(ParseError):
        parser.parse('a b')



import pytest
import os
from parglare.parser import Parser
from parglare.exceptions import ParseError
from .expression_grammar import get_grammar


def test_invalid_input():

    grammar = get_grammar()
    p = Parser(grammar)

    try:
        p.parse("id+id*+id")
    except ParseError as e:
        assert e.position == 6
        symbol_names = [s.name for s in e.symbols]
        assert "(" in symbol_names
        assert "id" in symbol_names
        assert len(symbol_names) == 2


def test_premature_end():

    grammar = get_grammar()
    p = Parser(grammar)

    try:
        p.parse("id+id*")
    except ParseError as e:
        assert e.position == 6
        symbol_names = [s.name for s in e.symbols]
        assert "(" in symbol_names
        assert "id" in symbol_names
        assert len(symbol_names) == 2


def test_line_column():
    grammar = get_grammar()
    p = Parser(grammar)

    try:
        p.parse("""id + id * id + id + error * id""")
    except ParseError as e:
        assert e.position == 20
        assert e.line == 1
        assert e.column == 20

    try:
        p.parse("""id + id * id + id + error * id

        """)
    except ParseError as e:
        assert e.position == 20
        assert e.line == 1
        assert e.column == 20

    try:
        p.parse("""

id + id * id + id + error * id""")
    except ParseError as e:
        assert e.position == 22
        assert e.line == 3
        assert e.column == 20

    try:
        p.parse("""

id + id * id + id + error * id

        """)
    except ParseError as e:
        assert e.position == 22
        assert e.line == 3
        assert e.column == 20


def test_file_name():
    "Test that file name is given in the error string when parsing file."
    grammar = get_grammar()
    p = Parser(grammar)

    input_file = os.path.join(os.path.dirname(__file__),
                              'parsing_errors.txt')

    try:
        p.parse_file(input_file)
    except ParseError as e:
        assert 'parsing_errors.txt' in str(e)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function
import codecs
from .grammar import Grammar, TerminalStr, EMPTY, AUGSYMBOL, EOF, STOP
from .exceptions import ParseError

SHIFT = 0
REDUCE = 1
ACCEPT = 2

# Tables construction algorithms
SLR = 0
LALR = 1


class Parser(object):
    """Parser works like a DFA driven by LR tables. For a given grammar LR table
    will be created and cached or loaded from cache if cache is found.
    """
    def __init__(self, grammar, start_production=1, actions=None, debug=False,
                 layout_debug=False, ws='\n\t ', default_actions=True,
                 tables=LALR, layout=False, position=False):
        self.grammar = grammar
        self.start_production = start_production
        self.actions = actions if actions else {}

        self.layout_parser = None
        if not layout:
            layout_prod = grammar.get_production('LAYOUT')
            if layout_prod:
                self.layout_parser = Parser(grammar,
                                            start_production=layout_prod,
                                            ws=None, layout=True,
                                            position=True,
                                            debug=layout_debug)

        self.layout = layout
        self.ws = ws
        self.position = position
        self.debug = debug
        self.layout_debug = layout_debug
        self.file_name = None

        self._states = []
        self._actions = []
        self._goto = []

        self.default_actions = default_actions

        from .closure import LR_0, LR_1
        from .tables import create_tables
        if tables == SLR:
            itemset_type = LR_0
        else:
            itemset_type = LR_1
        create_tables(self, itemset_type)

    def print_debug(self):
        if self.layout and self.layout_debug:
            print('\n\n*** LAYOUT parser ***\n')
        self.grammar.print_debug()
        print("\n\n*** STATES ***")
        for s, g, a in zip(self._states, self._goto, self._actions):
            s.print_debug()

            if g:
                print("\n\n\tGOTO:")
                print("\t", ", ".join(["%s->%d" % (k, v.state_id)
                                       for k, v in g.items()]))
            print("\n\tACTIONS:")
            print("\t", ", ".join(["%s->%s" % (k, str(v))
                                   for k, v in a.items()]))

    def lexical_disambiguation(self, tokens):
        """For the given list of matched tokens apply disambiguation strategy.
        Args:
        tokens (list of tuples (Terminal, matched str))
        """

        # By priority
        tokens.sort(key=lambda x: x[0].prior, reverse=True)
        prior = tokens[0][0].prior
        tokens = [x for x in tokens if x[0].prior == prior]
        if len(tokens) == 1:
            return tokens[0]

        # Multiple with the same priority. Use longest-match rule.
        tokens.sort(key=lambda x: len(x[1]), reverse=True)
        max_len = len(tokens[0][1])
        tokens = [x for x in tokens if len(x[1]) == max_len]
        if len(tokens) == 1:
            return tokens[0]

        # Multiple with the same length. Use most-specific rule.
        tokens_str = [x for x in tokens if isinstance(x[0], TerminalStr)]
        if len(tokens_str) > 0:
            # In any case return the first one
            return tokens_str[0]

        # If all fails return the first with the longest match.
        return tokens[0]

    def parse_file(self, file_name):
        """
        Parses content from the given file.
        Args:
            file_name(str): A file name.
        """
        self.file_name = file_name
        with codecs.open(file_name, 'r', 'utf-8') as f:
            content = f.read()
        return self.parse(content)

    def parse(self, input_str, position=0):
        """
        Parses the given input string.
        Args:
            input_str(str): A string to parse.
            position(int): Position to start from.
        """

        if self.debug:
            print("*** Parsing started")

        state_stack = [self._states[0]]
        results_stack = []
        position_stack = [0]
        position = position
        in_len = len(input_str)

        context = type(str("Context"), (), {})

        new_token = True
        ntok_sym = None

        while True:
            cur_state = state_stack[-1]
            if self.debug:
                print("Current state =", cur_state.state_id)
            goto = self._goto[cur_state.state_id]
            actions = self._actions[cur_state.state_id]

            if new_token or ntok_sym not in actions:
                # Try to recognize a new token in the input only after
                # successful SHIFT operation, i.e. when the input position has
                # moved. REDUCE operation doesn't move position. If the current
                # ntok_sym is not in the current actions then try to find new
                # token. This could happen if old token is not available in the
                # actions of the current state but EMPTY is and will match
                # always leading to reduction.

                # Parse layout
                layout_content = ''
                if self.layout_parser:
                    layout_content, pos = self.layout_parser.parse(
                        input_str, position)
                    layout_content = input_str[position:pos]
                    context.layout = layout_content
                    position = pos
                elif self.ws:
                    old_pos = position
                    while position < in_len and input_str[position] in self.ws:
                        position += 1
                    layout_content = input_str[old_pos:position]
                    context.layout = layout_content

                if self.debug:
                    print("\tLayout content: '{}'".format(layout_content))

                # Find the next token in the input
                ntok = ''
                if position == in_len and EMPTY not in actions \
                   and STOP not in actions:
                    ntok_sym = EOF
                else:
                    tokens = []
                    for symbol in actions:
                        tok = symbol.parse(input_str, position)
                        if tok:
                            tokens.append((symbol, tok))
                    if not tokens:
                        if STOP in actions:
                            ntok_sym = STOP
                        else:
                            ntok_sym = EMPTY
                    elif len(tokens) == 1:
                        ntok_sym, ntok = tokens[0]
                    else:
                        ntok_sym, ntok = self.lexical_disambiguation(tokens)

            if self.debug:
                print("\tToken ahead:", ntok_sym)

            act = actions.get(ntok_sym)

            if not act:
                raise ParseError(self.file_name, input_str, position,
                                 actions.keys())

            context.position = position

            if act.action is SHIFT:
                state = act.state
                context.symbol = state.symbol

                if self.debug:
                    print("\tShift:{} \"{}\"".format(state.state_id, ntok),
                          "at position",
                          pos_to_line_col(input_str, position),
                          "=>", position_context(input_str, position))

                res = None
                if state.symbol.name in self.actions:
                    res = self.actions[state.symbol.name](context, ntok)
                elif self.default_actions:
                    res = default_shift_action(context, ntok)

                if self.debug:
                    print("\tAction result =", str(res))

                state_stack.append(state)
                results_stack.append(res)
                position_stack.append(position)
                position += len(ntok)
                new_token = True

            elif act.action is REDUCE:
                production = act.prod
                context.symbol = act.prod.symbol

                if self.debug:
                    print("\tReducing by prod '%s'." % str(production))

                subresults = results_stack[-len(production.rhs):]
                del state_stack[-len(production.rhs):]
                del results_stack[-len(production.rhs):]
                if len(production.rhs) > 1:
                    del position_stack[-(len(production.rhs)-1):]
                cur_state = state_stack[-1]
                goto = self._goto[cur_state.state_id]
                context.position = position_stack[-1]

                # Calling semantic actions
                res = None
                if act.prod.prod_symbol_id in self.actions:
                    res = self.actions[
                        act.prod.prod_symbol_id](context, subresults)
                elif act.prod.symbol.name in self.actions:
                    res = self.actions[
                        act.prod.symbol.name](context, subresults)
                elif self.default_actions:
                    res = default_reduce_action(context, nodes=subresults)

                if self.debug:
                    print("\tAction result =", str(res))

                state_stack.append(goto[production.symbol])
                results_stack.append(res)
                new_token = False

            elif act.action is ACCEPT:
                if self.debug:
                    print("SUCCESS!!!")
                assert len(results_stack) == 1
                if self.position:
                    return results_stack[0], position
                else:
                    return results_stack[0]


class Action(object):
    __slots__ = ['action', 'state', 'prod']

    def __init__(self, action, state=None, prod=None):
        self.action = action
        self.state = state
        self.prod = prod

    def __str__(self):
        ac = {SHIFT: 'SHIFT',
              REDUCE: 'REDUCE',
              ACCEPT: 'ACCEPT'}.get(self.action)
        if self.action == SHIFT:
            p = self.state.state_id
        elif self.action == REDUCE:
            p = self.prod.prod_id
        else:
            p = ''
        return '%s%s' % (ac, ':%d' % p if p else '')


class LRItem(object):
    __slots__ = ('production', 'position', 'follow')

    def __init__(self, production, position, follow=None):
        self.production = production
        self.position = position
        self.follow = set() if not follow else follow

    def __eq__(self, other):
        return other and self.production == other.production and \
            self.position == other.position

    def __repr__(self):
        return str(self)

    def __str__(self):
        s = []
        for idx, r in enumerate(self.production.rhs):
            if idx == self.position:
                s.append(".")
            s.append(str(r))
        if len(self.production.rhs) == self.position:
            s.append(".")
        s = " ".join(s)

        return "%d: %s = %s   follow=%s" % (self.production.prod_id,
                                            self.production.symbol, s,
                                            self.follow)

    @property
    def is_kernel(self):
        """
        Kernel items are items whose position is not at the beginning.
        The only exception to this rule is start symbol of the augmented
        grammar.
        """
        return self.position > 0 or self.production.symbol is AUGSYMBOL

    def get_pos_inc(self):
        """
        Returns new LRItem with incremented position or None if position
        cannot be incremented (e.g. it is already at the end of the production)
        """

        if self.position < len(self.production.rhs):
            return LRItem(self.production, self.position+1, self.follow)

    @property
    def is_at_end(self):
        """
        Is this items at the end position, e.g. a candidate for reduction.
        """
        return self.position == len(self.production.rhs)


class LRState(object):
    """LR State is a set of LR items."""
    __slots__ = ['parser', 'state_id', 'symbol', 'items',
                 '_per_next_symbol', '_max_prior_per_symbol']

    def __init__(self, parser, state_id, symbol, items):
        self.parser = parser
        self.state_id = state_id
        self.symbol = symbol
        self.items = items if items else []

    def __eq__(self, other):
        """Two states are equal if their kernel items are equal."""
        this_kernel = [x for x in self.items if x.is_kernel]
        other_kernel = [x for x in other.items if x.is_kernel]
        if len(this_kernel) != len(other_kernel):
            return False
        for item in this_kernel:
            if item not in other_kernel:
                return False
        return True

    @property
    def kernel_items(self):
        """
        Returns kernel items of this state.
        """
        return [i for i in self.items if i.is_kernel]

    def get_item(self, other_item):
        """
        Get this state item that is equal to the given other_item.
        """
        return self.items[self.items.index(other_item)]

    def print_debug(self):
        print("\nState %d" % self.state_id)
        for i in self.items:
            print("\t", i)


class Node(object):
    """A node of the parse tree.
    """
    __slots__ = ['position', 'symbol']

    def __init__(self, position, symbol):
        self.position = position
        self.symbol = symbol


class NodeNonTerm(Node):
    __slots__ = ['nodes']

    def __init__(self, position, symbol, nodes):
        super(NodeNonTerm, self).__init__(position, symbol)
        self.nodes = nodes

    def tree_str(self, depth=0):
        indent = '  ' * depth
        s = '{}[{}]'.format(self.symbol, self.position)
        if self.nodes:
            for n in self.nodes:
                if hasattr(n, 'tree_str'):
                    s += '\n' + indent + n.tree_str(depth+1)
                else:
                    s += '\n' + indent + n.__class__.__name__ \
                         + '(' + str(n) + ')'
        return s

    def __str__(self):
        return '<Node({}, {})>'.format(self.position, self.symbol)


class NodeTerm(Node):
    __slots__ = ['value']

    def __init__(self, position, symbol, value):
        super(NodeTerm, self).__init__(position, symbol)
        self.value = value

    def tree_str(self, depth=0):
        return '{}[{}, {}]'.format(self.symbol, self.position,
                                   self.value)

    def __str__(self):
        return '<Node({}, {}, {})>'.format(self.position,
                                           self.symbol, self.value)


def default_shift_action(context, value):
    return NodeTerm(context.position, context.symbol, value)


def default_reduce_action(context, nodes):
    return NodeNonTerm(context.position, context.symbol, nodes)


def first(grammar):
    """Calculates the sets of terminals that can start the sentence derived from
    all grammar symbols.
    """
    assert isinstance(grammar, Grammar), \
        "grammar parameter should be Grammar instance."

    first_sets = {}
    for t in grammar.terminals:
        first_sets[t] = set([t])

    def _first(nt):
        if nt in first_sets:
            return first_sets[nt]
        fs = set()
        first_sets[nt] = fs
        for p in grammar.productions:
            if p.symbol == nt:
                pfs = set()
                for r in p.rhs:
                    if r == nt:
                        pfs.discard(EMPTY)
                        break
                    f = _first(r)
                    pfs.update(f)
                    if EMPTY not in f:
                        pfs.discard(EMPTY)
                        break
                fs.update(pfs)
        return fs

    for nt in grammar.nonterminals:
        first_sets[nt] = _first(nt)

    return first_sets


def follow(grammar, first_sets=None):
    """Calculates the sets of terminals that can follow some non-terminal for the
    given grammar.

    Args:
    grammar (Grammar): An initialized grammar.
    first_sets (dict): A sets of FIRST terminals keyed by a grammar symbol.
    """

    if first_sets is None:
        first_sets = first(grammar)

    follow_sets = {}
    for symbol in grammar.nonterminals:
        follow_sets[symbol] = set()

    has_additions = True
    while has_additions:
        has_additions = False
        for symbol in grammar.nonterminals:
            for p in grammar.productions:
                for idx, s in enumerate(p.rhs):
                    if s == symbol:
                        prod_follow = set()
                        for rsymbol in p.rhs[idx+1:]:
                            sfollow = first_sets[rsymbol]
                            prod_follow.update(sfollow)
                            if EMPTY not in sfollow:
                                break
                        else:
                            prod_follow.update(follow_sets[p.symbol])
                        prod_follow.discard(EMPTY)
                        if prod_follow.difference(follow_sets[symbol]):
                            has_additions = True
                            follow_sets[symbol].update(prod_follow)
    return follow_sets


def pos_to_line_col(input_str, position):
    """
    Returns position in the (line,column) form.
    """
    line = 1
    old_pos = 0
    try:
        cur_pos = input_str.index("\n")
        while cur_pos < position:
            line += 1
            old_pos = cur_pos + 1
            cur_pos = input_str.index("\n", cur_pos + 1)
    except ValueError:
        pass

    return line, position - old_pos


def position_context(input_str, position):
    """
    Returns position context string.
    """
    start = max(position-10, 0)
    c = input_str[start:position] + "*" \
        + input_str[position:position+10]
    c = c.replace("\n", "\\n")
    return c



#-*- coding: utf-8 -*-
#######################################################################
# Testing parsing speed. This is used for the purpose of testing
#   of performance gains/loses for various approaches.
# Author: Igor R. Dejanovic <igor DOT dejanovic AT gmail DOT com>
# Copyright: (c) 2017 Igor R. Dejanovic <igor DOT dejanovic AT gmail DOT com>
# License: MIT License
#######################################################################
from __future__ import print_function, unicode_literals

import time
from os.path import dirname, join, getsize
from parglare import Parser, Grammar


def timeit(file_name, message, **kwargs):
    print(message, 'File:', file_name)
    file_name = join(dirname(__file__), 'test_inputs', file_name)
    file_size = getsize(file_name)
    print('File size: {:.2f}'.format(file_size/1000), 'KB')

    this_folder = dirname(__file__)
    g = Grammar.from_file(join(this_folder, 'rhapsody.pg'))
    parser = Parser(g, **kwargs)

    t_start = time.time()
    with open(file_name) as f:
        parser.parse(f.read())
    t_end = time.time()

    print('Elapsed time: {:.2f}'.format(t_end - t_start), 'sec')
    print('Speed = {:.2f}'.format(file_size/1000/(t_end - t_start)),
          'KB/sec\n')


def main():

    # Small file
    file_name_small = 'LightSwitch.rpy'
    # Large file
    file_name_large = 'LightSwitchDouble.rpy'

    # Without semantic actions
    for i in range(3):
        timeit(file_name_small,
               '{}. Small file without sem. actions.'.format(i + 1),
               default_actions=False)
        timeit(file_name_large,
               '{}. Large file without sem. actions.'.format(i + 1),
               default_actions=False)

    # With default semantic actions.
    for i in range(3):
        timeit(file_name_small,
               '{}. Small file with default semantic actions.'.format(i + 1),
               default_actions=True)
        timeit(file_name_large,
               '{}. Large file with default semantic actions.'.format(i + 1),
               default_actions=True)


if __name__ == '__main__':
    main()



class GraphStructuredStack(object):
    """
    A data structure representing a forkable stack from Tomita's paper.
    """


class PackedParseForestNode(object):
    """This is a node in the Tomita's packed parse forest."""


class NonTerminal(PackedParseForestNode):
    pass


class Terminal(PackedParseForestNode):
    pass



# -*- coding: utf-8 -*-
# flake8: NOQA
from parglare.parser import Parser, LALR, SLR
from parglare.grammar import Grammar, NonTerminal, Terminal, TerminalStr, \
    TerminalRegEx, create_grammar, EMPTY, EOF

__author__ = """Igor R. Dejanovic"""
__email__ = 'igor DOT dejanovic AT gmail DOT com'
__version__ = '0.1'



import pytest
from parglare import Parser, Grammar


def test_layout_whitespaces():
    grammar = r"""
    S = K EOF;
    K = A | B;
    A = 'a' A | 'a';
    B = 'b' B | 'b';
    LAYOUT = WS | EMPTY;
    WS = /\s+/;
    """

    g = Grammar.from_string(grammar)
    parser = Parser(g)

    parser.parse("""aaa a    aaaa
    aa    aa a aaa

    aaa
    """)


def test_layout_simple_comments():
    grammar = r"""
    S = K EOF;
    K = A | B;
    A = 'a' A | 'a';
    B = 'b' B | 'b';
    LAYOUT = LayoutItem | LAYOUT LayoutItem;
    LayoutItem = WS | Comment | EMPTY;
    WS = /\s+/;
    Comment = /\/\/.*/;
    """

    g = Grammar.from_string(grammar)
    parser = Parser(g)

    parser.parse("""aaa a    aaaa
    aa    aa a aaa // This is a comment

    aaa
    """)


def test_layout_nested_comments():
    grammar = """
    S = K EOF;
    K = 'a' B | 'a' C;
    B = 'b' | B 'b';
    C = 'c' | C 'c';

    LAYOUT = LayoutItem | LAYOUT LayoutItem;
    LayoutItem = WS | Comment | EMPTY;
    WS = /\s+/;
    Comment = '/*' CorNCs '*/' | /\/\/.*/;
    CorNCs = CorNC | CorNCs CorNC | EMPTY;
    CorNC = Comment | NotComment | WS;
    NotComment = /((\*[^\/])|[^*\/]|\/[^\*])+/;
    """
    g = Grammar.from_string(grammar)
    parser = Parser(g)
    parser.parse("""
    a  b b b   b // This is line comment
    b b b b b b  /* This is block
    comment */

    bbbb  b b b b b
    /* Another block comment
       // With nested line comment
       /* And nested block
    comment */
    */

    bbbb b b b
    """)


def test_layout_context():
    """
    Test that layout is passed in the action context.
    """
    grammar = r"""
    S = K EOF;
    K = A | B;
    A = 'a' A | 'a';
    B = 'b' B | 'b';
    LAYOUT = LayoutItem | LAYOUT LayoutItem;
    LayoutItem = WS | Comment | EMPTY;
    WS = /\s+/;
    Comment = /\/\/.*/;
    """

    layout_called = [False]
    layout_passed = [False]

    def layout_action(context, _):
        layout_called[0] = True
        if 'This is a comment' in context.layout:
            layout_passed[0] = True

    actions = {
        "a": layout_action
    }

    g = Grammar.from_string(grammar)
    parser = Parser(g, actions=actions)

    parser.parse("""aaa a    aaaa
    aa    aa a aaa // This is a comment

    aaa
    """)

    assert layout_called[0]
    assert layout_passed[0]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import sys
from setuptools import setup
try:
    import pypandoc
    readme = pypandoc.convert('README.md', 'rst')
    history = pypandoc.convert('HISTORY.md', 'rst')
except(IOError, ImportError):
    readme = open('README.md').read()
    history = open('HISTORY.md').read()


VERSION = '0.1'

if sys.argv[-1].startswith('publish'):
    if os.system("pip list | grep wheel"):
        print("wheel not installed.\nUse `pip install wheel`.\nExiting.")
        sys.exit()
    if os.system("pip list | grep twine"):
        print("twine not installed.\nUse `pip install twine`.\nExiting.")
        sys.exit()
    os.system("python setup.py sdist bdist_wheel")
    if sys.argv[-1] == 'publishtest':
        os.system("twine register -r test dist/parglare-{}.tar.gz"
                  .format(VERSION))
        os.system(
            "twine register -r test dist/parglare-{}-py2.py3-none-any.whl"
            .format(VERSION))
        os.system("twine upload -r test dist/*")
    else:
        os.system("twine register dist/parglare-{}.tar.gz"
                  .format(VERSION))
        os.system(
            "twine register dist/parglare-{}-py2.py3-none-any.whl"
            .format(VERSION))
        os.system("twine upload dist/*")
        print("You probably want to also tag the version now:")
        print("  git tag -a {0} -m 'version {0}'".format(VERSION))
        print("  git push --tags")
    sys.exit()

requirements = [
    # TODO: put package requirements here
]

test_requirements = [
    # TODO: put package test requirements here
]

setup(
    name='parglare',
    version=VERSION,
    description="A pure Python Scannerless LR (soon GLR) parser ",
    long_description=readme + '\n\n' + history,
    author="Igor R. Dejanovic",
    author_email='igor DOT dejanovic AT gmail DOT com',
    url='https://github.com/igordejanovic/parglare',
    packages=[
        'parglare',
    ],
    package_dir={'parglare':
                 'parglare'},
    include_package_data=True,
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords='parglare',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'Topic :: Software Development :: Interpreters',
        'Topic :: Software Development :: Compilers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Operating System :: OS Independent',
    ],
    test_suite='tests',
    tests_require=test_requirements
)



from parglare import create_grammar
from parglare import NonTerminal, TerminalStr, TerminalRegEx


# Expression grammar with float numbers
E, T, F = [NonTerminal(name) for name in ['E', 'T', 'F']]
PLUS, MULT, OPEN, CLOSE = [
    TerminalStr(value, value) for value in ['+', '*', '(', ')']]
NUMBER = TerminalRegEx('number', r'\d+(\.\d+)?')
productions = [
    (E, (E, PLUS, T)),
    (E, (T, )),
    (T, (T, MULT, F)),
    (T, (F, )),
    (F, (OPEN, E, CLOSE)),
    (F, (NUMBER,))
]


def get_grammar():
    return create_grammar(productions, E)



class GrammarError(Exception):
    pass


class ParseError(Exception):
    def __init__(self, file_name, input_str, position, symbols):
        from parglare.parser import pos_to_line_col, position_context
        self.file_name = file_name
        self.position = position
        self.symbols = symbols
        self.line, self.column = pos_to_line_col(input_str, position)
        context = position_context(input_str, position)
        super(ParseError, self).__init__(
            'Error {}at position {},{} => "{}". Expected: {}'.format(
                'in file "{}" '.format(self.file_name)
                if self.file_name else "",
                self.line, self.column, context,
                ' or '.join([s.name for s in symbols])))


class ParserInitError(Exception):
    pass


class LRConflict(Exception):
    def __init__(self, message, state, symbol):
        self.state = state
        self.symbol = symbol
        super(LRConflict, self).__init__(message)


class ShiftReduceConflict(LRConflict):
    def __init__(self, state, symbol, production):
        self.production = production
        message = "In state {} and input symbol '{}' can't " \
                  "decide whether to shift or reduce by production '{}'." \
            .format(state.state_id, symbol, production)
        super(ShiftReduceConflict, self).__init__(message, state, symbol)


class ReduceReduceConflict(LRConflict):
    def __init__(self, state, symbol, production1, production2):
        self.production1 = production1
        self.production2 = production2
        message = "In state {} and input symbol '{}' can't " \
                  "decide whether to reduce by production '{}' or by '{}'." \
            .format(state.state_id, symbol, production1, production2)
        super(ReduceReduceConflict, self).__init__(message, state, symbol)
